# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-03-04 08:42
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('joyous', '0002_auto_20180228_1449'),
    ]

    operations = [
        migrations.AddField(
            model_name='extrainfopage',
            name='extra_title',
            field=models.CharField(blank=True, help_text='A more specific title for this occurence (optional)', max_length=255, verbose_name='Title'),
        ),
    ]



# ------------------------------------------------------------------------------
# ical import/export format
# ------------------------------------------------------------------------------
import datetime as dt
import pytz
from socket import gethostname
from icalendar import Calendar, Event
from icalendar import vDatetime
#from html2text import HTML2Text
from django.http import HttpResponse
from django.utils import timezone
from ls.joyous import __version__
from ..models import (SimpleEventPage, MultidayEventPage, RecurringEventPage,
        EventExceptionBase, ExtraInfoPage, CancellationPage)
from ..utils.telltime import getAwareDatetime
from .vtimezone import create_timezone

# ------------------------------------------------------------------------------
class ICalendarHander:
    def serve(self, page, request, *args, **kwargs):
        vcal = self.makeVCalendar(page)
        if vcal is not None:
            response = HttpResponse(vcal.export(),
                                    content_type='text/calendar')
            response['Content-Disposition'] = \
                'attachment; filename={}.ics'.format(page.slug)
            return response
        return None

    def makeVCalendar(self, page):
        vcal = None
        vevent = self.makeVEvent(page)
        if vevent is not None:
            vcal = VCalendar()
            if page.tz and page.tz is not pytz.utc:
                vtz = create_timezone(page.tz, vevent.dtstart, vevent.dtend)
                vcal.add_component(vtz)
            vcal.add_component(vevent)
            for vchild in vevent.vchildren:
                vcal.add_component(vchild)
        return vcal

    def makeVEvent(self, page):
        if isinstance(page, SimpleEventPage):
            return SimpleVEvent(page)
        elif isinstance(page, MultidayEventPage):
            return MultidayVEvent(page)
        elif isinstance(page, RecurringEventPage):
            return RecurringVEvent(page)
        elif isinstance(page, EventExceptionBase):
            return RecurringVEvent(page.overrides)
        else:
            return None

# ------------------------------------------------------------------------------
class VCalendar(Calendar):
    prodVersion = ".".join(__version__.split(".", 2)[:2])
    prodId = "-//linuxsoftware.nz//NONSGML Joyous v{}//EN".format(prodVersion)

    def __init__(self):
        super().__init__()
        self.add('prodid',  self.prodId)
        self.add('version', "2.0")

    def export(self):
        return self.to_ical()

# ------------------------------------------------------------------------------
class VEvent(Event):
    def __init__(self, page):
        super().__init__()
        #h2t = HTML2Text()
        self.page = page
        self.vchildren = []
        firstRevision = page.revisions.order_by("created_at").first()
        self.add('UID',           self.uid)
        self.add('URL',           page.full_url)
        self.add('SUMMARY',       self.summary)
        self.add('DESCRIPTION',   self.description)
        self.add('SEQUENCE',      page.revisions.count())
        self.add('DTSTAMP',       vDatetime(timezone.now()))
        self.add('CREATED',       vDatetime(firstRevision.created_at))
        self.add('LAST-MODIFIED', vDatetime(page.latest_revision_created_at))
        self.add('LOCATION',      self.location)
        self.add('DTSTART',       vDatetime(self.dtstart))
        self.add('DTEND',         vDatetime(self.dtend))

    @property
    def uid(self):
        pg = self.page
        site = pg.get_site()
        hostname = site.hostname if site else gethostname()
        return "{}-{}@{}".format(pg.id, pg.slug, hostname)

    @property
    def summary(self):
        return self.page.title

    @property
    def description(self):
        #return h2t.handle(self.page.details)
        return self.page.details

    @property
    def location(self):
        return self.page.location


class SimpleVEvent(VEvent):
    def __init__(self, page):
        super().__init__(page)

    @property
    def dtstart(self):
        pg = self.page
        return getAwareDatetime(pg.date, pg.time_from, pg.tz, dt.time.min)

    @property
    def dtend(self):
        pg = self.page
        return getAwareDatetime(pg.date, pg.time_to, pg.tz, dt.time.max)


class MultidayVEvent(VEvent):
    def __init__(self, page):
        super().__init__(page)

    @property
    def dtstart(self):
        pg = self.page
        return getAwareDatetime(pg.date_from, pg.time_from, pg.tz, dt.time.min)

    @property
    def dtend(self):
        pg = self.page
        return getAwareDatetime(pg.date_to, pg.time_to, pg.tz, dt.time.max)


class RecurringVEvent(VEvent):
    def __init__(self, page):
        super().__init__(page)
        self.exDates = []
        self._handleExceptions()
        if self.exDates:
            self.add('EXDATE', self.exDates)
        self.add('RRULE', self._getRrule())

    @property
    def dtstart(self):
        startDt = self.page._getMyFirstDatetimeFrom()
        if startDt is not None:
            return startDt
        else:
            return pytz.utc.localize(dt.datetime.min)

    @property
    def dtend(self):
        endDt = self.page._getMyFirstDatetimeTo()
        if endDt is not None:
            return endDt
        else:
            return pytz.utc.localize(dt.datetime.min)

    def _handleExceptions(self):
        pg = self.page
        for cancellation in CancellationPage.objects.live().child_of(pg) \
                                            .select_related("postponementpage"):
            postponement = getattr(cancellation, "postponementpage", None)
            if not postponement and cancellation.cancellation_title == "":
                excludeDt = getAwareDatetime(cancellation.except_date,
                                             cancellation.time_from,
                                             cancellation.tz, dt.time.min)
                self.exDates.append(excludeDt)
            else:
                if cancellation.title != "":
                    self.vchildren.append(CancelledVEvent(cancellation, self))
                if postponement:
                    self.vchildren.append(PostponedVEvent(postponement, self))

        for info in ExtraInfoPage.objects.live().child_of(pg):
            self.vchildren.append(ExtraInfoVEvent(info, self))

    def _getRrule(self):
        repeat = self.page.repeat
        props = [('FREQ', repeat.frequency)]
        if repeat.interval and repeat.interval != 1:
            props.append(('INTERVAL', repeat.interval))
        if repeat.wkst:
            props.append(('WKST', repr(repeat.wkst)))
        if repeat.count:
            props.append(('COUNT', repeat.count))
        if repeat.until:
            props.append(('UNTIL', repeat.until))
        for name, value in [('BYSETPOS',   repeat.bysetpos),
                            ('BYDAY',      repeat.byweekday),
                            ('BYMONTH',    repeat.bymonth),
                            ('BYMONTHDAY', repeat.bymonthday),
                            ('BYYEARDAY',  repeat.byyearday),
                            ('BYWEEKNO',   repeat.byweekno)]:
            if value:
                props.append((name, [repr(v) for v in value]))
        return dict(props)


class ExceptionVEvent(VEvent):
    def __init__(self, page, vparent):
        self.vparent = vparent
        super().__init__(page)
        self.add('RECURRENCE-ID', vDatetime(self._getStartDt()))

    def _getStartDt(self):
        pg = self.page
        return getAwareDatetime(pg.except_date, pg.time_from, pg.tz, dt.time.min)

    def _getEndDt(self):
        pg = self.page
        return getAwareDatetime(pg.except_date, pg.time_to, pg.tz, dt.time.max)

    @property
    def uid(self):
        return self.vparent.uid

    @property
    def location(self):
        return self.vparent.page.location

    dtstart = property(_getStartDt)
    dtend   = property(_getEndDt)


class ExtraInfoVEvent(ExceptionVEvent):
    def __init__(self, page, vparent):
        super().__init__(page, vparent)

    @property
    def summary(self):
        return self.page.extra_title or self.vparent.page.title

    @property
    def description(self):
        return self.page.extra_information + "\n" + self.vparent.page.details


class CancelledVEvent(ExceptionVEvent):
    def __init__(self, page, vparent):
        super().__init__(page, vparent)

    @property
    def summary(self):
        return self.page.cancellation_title

    @property
    def description(self):
        return self.page.cancellation_details


class PostponedVEvent(ExceptionVEvent):
    def __init__(self, page, vparent):
        super().__init__(page, vparent)
        self.add('LOCATION', page.location)

    @property
    def summary(self):
        return self.page.postponement_title

    @property
    def location(self):
        return self.page.location

    @property
    def dtstart(self):
        pg = self.page
        return getAwareDatetime(pg.date, pg.time_from, pg.tz, dt.time.min)

    @property
    def dtend(self):
        pg = self.page
        return getAwareDatetime(pg.date, pg.time_to, pg.tz, dt.time.max)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Test Telltime Utilities
# ------------------------------------------------------------------------------
import sys
import datetime as dt
import pytz
from django.test import TestCase
from .testutils import datetimetz
from ls.joyous.utils.telltime import (getAwareDatetime, getLocalDatetime,
        getLocalDateAndTime, getLocalDate, getLocalTime,
        timeFrom, timeTo, timeFormat, dateFormat, dateFormatDMY)

class TestLocalTimes(TestCase):
    def testGetAwareDatetime(self):
        when = getAwareDatetime(dt.date(1999,12,1), dt.time(2),
                                pytz.timezone("Asia/Kuala_Lumpur"))
        self.assertEqual(when.tzinfo.zone, "Asia/Kuala_Lumpur")
        self.assertEqual(when.date(), dt.date(1999,12,1))
        self.assertEqual(when.time(), dt.time(2))
        when = getAwareDatetime(dt.date(2004,2,15), None,
                                pytz.timezone("Australia/Melbourne"))
        self.assertEqual(when.tzinfo.zone, "Australia/Melbourne")
        self.assertEqual(when.date(), dt.date(2004,2,15))
        self.assertEqual(when.time(), dt.time.max)

    def testGetLocalDatetime(self):
        self.assertEqual(getLocalDatetime(dt.date(2019,1,1), dt.time(1)),
                         datetimetz(2019,1,1,1))
        when = getLocalDatetime(dt.date(2003,9,2), dt.time(10,45,1),
                                pytz.timezone("Asia/Tokyo"))
        self.assertEqual(when, datetimetz(2003,9,2,10,45,1))
        when = getLocalDatetime(dt.date(2017,3,23), dt.time(18),
                                pytz.timezone("Europe/Prague"))
        self.assertEqual(when.tzinfo.zone, "Asia/Tokyo")
        self.assertEqual(when.date(), dt.date(2017,3,24))
        self.assertEqual(when.time(), dt.time(2)),
        when = getLocalDatetime(dt.date(2006,6,22), None,
                                pytz.timezone("America/Toronto"), dt.time(0))
        self.assertEqual(when.tzinfo.zone, "Asia/Tokyo")
        self.assertEqual(when.date(), dt.date(2006,6,22))
        self.assertEqual(when.time(), dt.time(0))

    def testGetLocalDateAndTime(self):
        date, time = getLocalDateAndTime(dt.date(1987,1,1), dt.time(1))
        self.assertEqual(date, dt.date(1987,1,1))
        localTZ = pytz.timezone("Asia/Tokyo")
        self.assertEqual(time, dt.time(1).replace(tzinfo=localTZ))
        date, time = getLocalDateAndTime(dt.date(2011,4,28), None,
                                pytz.timezone("Asia/Yerevan"))
        self.assertEqual(date, dt.date(2011,4,29))
        self.assertEqual(time, None)

    def testGetLocalDate(self):
        date = getLocalDate(dt.date(1993,8,8), None,
                            pytz.timezone("Europe/London"))
        self.assertEqual(date, dt.date(1993,8,9))

    def testGetLocalTime(self):
        time = getLocalTime(dt.date(2018,5,8), dt.time(22,44),
                            pytz.timezone("Pacific/Auckland"))
        localTZ = pytz.timezone("Asia/Tokyo")
        self.assertEqual(time, dt.time(19,44).replace(tzinfo=localTZ))


class TestNullableTimes(TestCase):
    def testTimeFrom(self):
        self.assertEqual(timeFrom(None), dt.time(0))
        self.assertEqual(timeFrom(dt.time(8)), dt.time(8))

    def testTimeTo(self):
        self.assertEqual(timeTo(None), dt.time.max)
        self.assertEqual(timeTo(dt.time(8)), dt.time(8))


class TestFormats(TestCase):
    def testTimeFormat(self):
        self.assertEqual(timeFormat(None), "")
        self.assertEqual(timeFormat("","","Aaa","Bbb"), "")
        self.assertEqual(timeFormat(dt.time(13,26)), "1:26pm")
        self.assertEqual(timeFormat(dt.time(8), prefix="~~"), "~~8am")
        self.assertEqual(timeFormat(dt.time(8), dt.time(11)), "8am to 11am")
        self.assertEqual(timeFormat(dt.time(20), dt.time(2), "at ", "-> "),
                         "at 8pm -> 2am")

    def testDateFormat(self):
        self.assertEqual(dateFormat(None), "")
        self.assertEqual(dateFormat(dt.date(2017,2,16)),
                         "Thursday 16th of February 2017")
        today = dt.date.today()
        parts = dateFormat(today).split()
        self.assertEqual(len(parts), 4)
        self.assertEqual(parts[0], "{:%A}".format(today))
        self.assertEqual(int(parts[1][:-2]), today.day)
        self.assertIn(parts[1][-2:], ["st", "nd", "rd", "th"])
        self.assertEqual(parts[2], "of")
        self.assertEqual(parts[3], "{:%B}".format(today))

    def testDateFormatDMY(self):
        self.assertEqual(dateFormatDMY(dt.date(2016,5,22)), "22 May 2016")
        self.assertEqual(dateFormatDMY(None), "")





# ------------------------------------------------------------------------------
# Test Multiday Event Page
# ------------------------------------------------------------------------------
import sys
import datetime as dt
import pytz
from django.test import TestCase
from django.contrib.auth.models import User
from django.utils import timezone
from wagtail.core.models import Page
from ls.joyous.models.calendar import CalendarPage
from ls.joyous.models.events import MultidayEventPage
from .testutils import datetimetz

class TestMultidayEvent(TestCase):
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@ok.test', 's3cr3t')
        self.calendar = CalendarPage(owner = self.user,
                                     slug  = "events",
                                     title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()
        self.event = MultidayEventPage(owner = self.user,
                                       slug  = "allnighter",
                                       title = "All Night",
                                       date_from = dt.date(2012,12,31),
                                       date_to   = dt.date(2013,1,1),
                                       time_from = dt.time(23),
                                       time_to   = dt.time(3))
        self.calendar.add_child(instance=self.event)
        self.event.save_revision().publish()

    def testGetEventsByDay(self):
        events = MultidayEventPage.getEventsByDay(dt.date(2012,12,1),
                                                  dt.date(2012,12,31))
        self.assertEqual(len(events), 31)
        evod = events[30]
        self.assertEqual(evod.date, dt.date(2012,12,31))
        self.assertEqual(len(evod.all_events), 1)
        self.assertEqual(len(evod.days_events), 1)
        self.assertEqual(len(evod.continuing_events), 0)
        events = MultidayEventPage.getEventsByDay(dt.date(2013,1,1),
                                                  dt.date(2013,1,31))
        self.assertEqual(len(events), 31)
        evod = events[0]
        self.assertEqual(evod.date, dt.date(2013,1,1))
        self.assertEqual(len(evod.all_events), 1)
        self.assertEqual(len(evod.days_events), 0)
        self.assertEqual(len(evod.continuing_events), 1)

    def testStatus(self):
        self.assertEqual(self.event.status, "finished")
        self.assertEqual(self.event.status_text, "This event has finished.")
        today = timezone.localdate()
        yesterday = today - dt.timedelta(1)
        nextWeek = today + dt.timedelta(6 - today.weekday())
        nowEvent = MultidayEventPage(owner = self.user,
                                     slug  = "now",
                                     title = "Now Event",
                                     date_from = yesterday,
                                     date_to   = nextWeek)
        self.calendar.add_child(instance=nowEvent)
        self.assertEqual(nowEvent.status, "started")
        self.assertEqual(nowEvent.status_text, "This event has started.")
        tomorrow = today + dt.timedelta(days=1)
        futureEvent = MultidayEventPage(owner = self.user,
                                        slug  = "tomorrow",
                                        title = "Tomorrow's Event",
                                        date_from  = tomorrow,
                                        date_to    = tomorrow + dt.timedelta(days=1))
        self.calendar.add_child(instance=futureEvent)
        self.assertIsNone(futureEvent.status)
        self.assertEqual(futureEvent.status_text, "")

    def testWhen(self):
        self.assertEqual(self.event.when, "Monday 31st of December 2012 at 11pm to "
                                          "Tuesday 1st of January 2013 at 3am")

    def testAt(self):
        self.assertEqual(self.event.at, "11pm")

    def testUpcomingDt(self):
        self.assertIsNone(self.event._upcoming_datetime_from)
        now = timezone.localtime()
        today = now.date()
        nextWeek = today + dt.timedelta(6 - today.weekday())
        earlier = now - dt.timedelta(hours=1)
        if earlier.date() != now.date():
            earlier = dt.datetime.combine(now.date(), dt.time.min)
        nowEvent = MultidayEventPage(owner = self.user,
                                     slug  = "now",
                                     title = "Now Event",
                                     date_from = earlier.date(),
                                     date_to   = nextWeek,
                                     time_from = earlier.time(),
                                     time_to   = dt.time(1))
        self.calendar.add_child(instance=nowEvent)
        self.assertIsNone(nowEvent._upcoming_datetime_from)
        tomorrow = timezone.localdate() + dt.timedelta(days=1)
        futureEvent = MultidayEventPage(owner = self.user,
                                        slug  = "tomorrow",
                                        title = "Tomorrow's Event",
                                        date_from  = tomorrow,
                                        date_to    = tomorrow + dt.timedelta(2))
        self.calendar.add_child(instance=futureEvent)
        self.assertEqual(futureEvent._upcoming_datetime_from,
                         datetimetz(tomorrow, dt.time.max))

    def testPastDt(self):
        self.assertEqual(self.event._past_datetime_from,
                         datetimetz(2012,12,31,23,0))
        now = timezone.localtime()
        today = now.date()
        nextWeek = today + dt.timedelta(6 - today.weekday())
        earlier = now - dt.timedelta(hours=1)
        if earlier.date() != now.date():
            earlier = dt.datetime.combine(now.date(), dt.time.min)
        nowEvent = MultidayEventPage(owner = self.user,
                                     slug  = "now",
                                     title = "Now Event",
                                     date_from = earlier.date(),
                                     date_to   = nextWeek,
                                     time_from = earlier.time(),
                                     time_to   = dt.time(1))
        self.calendar.add_child(instance=nowEvent)
        self.assertEqual(nowEvent._past_datetime_from, earlier)
        tomorrow = timezone.localdate() + dt.timedelta(days=1)
        futureEvent = MultidayEventPage(owner = self.user,
                                        slug  = "tomorrow",
                                        title = "Tomorrow's Event",
                                        date_from  = tomorrow,
                                        date_to    = tomorrow + dt.timedelta(2))
        self.calendar.add_child(instance=futureEvent)
        self.assertIsNone(futureEvent._past_datetime_from)

    def testGroup(self):
        self.assertIsNone(self.event.group)


class TestMultidayEventTZ(TestCase):
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@ok.test', 's3cr3t')
        self.calendar = CalendarPage(owner = self.user,
                                     slug  = "events",
                                     title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()
        self.event = MultidayEventPage(owner = self.user,
                                       slug  = "niuekulele2018",
                                       title = "Niuekulele Ukulele Music Festival",
                                       date_from = dt.date(2018,3,16),
                                       date_to   = dt.date(2018,3,20),
                                       tz = pytz.timezone("Pacific/Niue"))
        self.calendar.add_child(instance=self.event)
        self.event.save_revision().publish()

    def testGetEventsByLocalDay(self):
        events = MultidayEventPage.getEventsByDay(dt.date(2018,3,1),
                                                  dt.date(2018,3,31))
        self.assertEqual(len(events), 31)
        evod1 = events[16]
        self.assertEqual(evod1.date, dt.date(2018,3,17))
        self.assertEqual(len(evod1.days_events), 1)
        self.assertEqual(len(evod1.continuing_events), 0)
        evod5 = events[20]
        self.assertEqual(evod5.date, dt.date(2018,3,21))
        self.assertEqual(len(evod5.days_events), 0)
        self.assertEqual(len(evod5.continuing_events), 1)
        self.assertEqual(evod1.all_events[0], evod5.all_events[0])
        self.assertIs(evod1.all_events[0].page, evod5.all_events[0].page)

    def testLocalWhen(self):
        self.assertEqual(self.event.when,
                         "Friday 16th of March to Wednesday 21st of March")

    def testLocalAt(self):
        self.assertEqual(self.event.at, "")

    @timezone.override("America/Los_Angeles")
    def testUpcomingLocalDt(self):
        self.assertIsNone(self.event._upcoming_datetime_from)

    @timezone.override("Pacific/Auckland")
    def testPastLocalDt(self):
        when = self.event._past_datetime_from
        self.assertEqual(when.tzinfo.zone, "Pacific/Auckland")
        self.assertEqual(when.time(), dt.time.max)
        self.assertEqual(when.date(), dt.date(2018,3,17))



# ------------------------------------------------------------------------------
# Joyous template tags
# ------------------------------------------------------------------------------
import datetime as dt
import calendar
from django import template
from ..utils.telltime import timeFormat, dateFormat
from ..models import getAllEventsByDay
from ..models import getAllUpcomingEvents
from ..models import getGroupUpcomingEvents
from ..models import getAllEventsByWeek
from ..models import CalendarPage
from ..utils.weeks import weekday_abbr, weekday_name

register = template.Library()

@register.inclusion_tag("joyous/tags/events_this_week.html")
def events_this_week():
    today = dt.date.today()
    begin_ord = today.toordinal()
    if today.weekday() != 6:
        # Start week with Monday, unless today is Sunday
        begin_ord -= today.weekday()
    end_ord = begin_ord + 6
    date_from = dt.date.fromordinal(begin_ord)
    date_to   = dt.date.fromordinal(end_ord)
    events = getAllEventsByDay(date_from, date_to)
    return {'events': events, 'today':  today }

@register.inclusion_tag("joyous/tags/minicalendar.html",
                        takes_context=True)
def minicalendar(context):
    today = dt.date.today()
    request = context['request']
    home = request.site.root_page
    cal = CalendarPage.objects.live().descendant_of(home).first()
    calUrl = cal.get_url(request) if cal else None
    return {'today':           today,
            'year':            today.year,
            'month':           today.month,
            'calendarUrl':     calUrl,
            'monthName':       calendar.month_name[today.month],
            'weekdayInfo':     zip(weekday_abbr, weekday_name),
            'events':          getAllEventsByWeek(today.year, today.month)}

@register.inclusion_tag("joyous/tags/upcoming_events_detailed.html")
def all_upcoming_events():
    return {'events': getAllUpcomingEvents()}

@register.inclusion_tag("joyous/tags/upcoming_events_detailed.html",
                        takes_context=True)
def subsite_upcoming_events(context):
    home = context['request'].site.root_page
    return {'events': getAllUpcomingEvents(home)}

@register.inclusion_tag("joyous/tags/upcoming_events_list.html",
                        takes_context=True)
def group_upcoming_events(context):
    page = context.get('page')
    events = getGroupUpcomingEvents(page) if page is not None else []
    return {'events': events}

# Format times and dates e.g. on event page
@register.filter
def time_display(time):
    return timeFormat(time)

@register.filter
def at_time_display(time):
    return timeFormat(time, prefix="at ")

@register.filter
def date_display(date):
    return dateFormat(date)




# ------------------------------------------------------------------------------
# Test Cancellation Page
# ------------------------------------------------------------------------------
import sys
import datetime as dt
from django.test import TestCase
from django.contrib.auth.models import User
from wagtail.core.models import Page
from ls.joyous.models.calendar import CalendarPage
from ls.joyous.models.events import RecurringEventPage
from ls.joyous.models.events import CancellationPage
from ls.joyous.recurrence import Recurrence, WEEKLY, MO, WE, FR


class TestCancellation(TestCase):
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@ok.test', 's3(r3t')
        self.calendar = CalendarPage(owner = self.user,
                                     slug  = "events",
                                     title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()
        self.event = RecurringEventPage(slug      = "test-meeting",
                                        title     = "Test Meeting",
                                        repeat    = Recurrence(dtstart=dt.date(1989,1,1),
                                                               freq=WEEKLY,
                                                               byweekday=[MO,WE,FR]),
                                        time_from = dt.time(13),
                                        time_to   = dt.time(15,30))
        self.calendar.add_child(instance=self.event)
        self.cancellation = CancellationPage(owner = self.user,
                                             slug  = "1989-02-01-cancellation",
                                             title = "Cancellation for Wednesday 1st of February",
                                             overrides = self.event,
                                             except_date = dt.date(1989,2,1),
                                             cancellation_title   = "Meeting Cancelled",
                                             cancellation_details =
                                                 "Cancelled due to lack of interest")
        self.event.add_child(instance=self.cancellation)
        self.cancellation.save_revision().publish()

    def testGetEventsByDay(self):
        hiddenCancellation = CancellationPage(owner = self.user,
                                              slug  = "1989-02-13-cancellation",
                                              title = "Cancellation for Monday 13th of February",
                                              overrides = self.event,
                                              except_date = dt.date(1989,2,13))
        self.event.add_child(instance=hiddenCancellation)
        events = RecurringEventPage.getEventsByDay(dt.date(1989,2,1),
                                                   dt.date(1989,2,28))
        self.assertEqual(len(events), 28)
        evod1 = events[0]
        self.assertEqual(evod1.date, dt.date(1989,2,1))
        self.assertEqual(len(evod1.days_events), 1)
        self.assertEqual(len(evod1.continuing_events), 0)
        title, page = evod1.days_events[0]
        self.assertEqual(title, "Meeting Cancelled")
        self.assertIs(type(page), CancellationPage)
        evod2 = events[12]
        self.assertEqual(evod2.date, dt.date(1989,2,13))
        self.assertEqual(len(evod2.days_events), 0)
        self.assertEqual(len(evod2.continuing_events), 0)

    def testStatus(self):
        self.assertEqual(self.cancellation.status, "cancelled")
        self.assertEqual(self.cancellation.status_text, "This event has been cancelled.")
        now = dt.datetime.now()
        myday = now.date() + dt.timedelta(1)
        friday = myday + dt.timedelta(days=(4-myday.weekday())%7)
        futureCan = CancellationPage(owner = self.user,
                                     slug  = "fri-cancellation",
                                     title = "Cancellation for Friday",
                                     overrides = self.event,
                                     except_date = friday,
                                     cancellation_title   = "",
                                     cancellation_details = "")
        self.event.add_child(instance=futureCan)
        self.assertEqual(futureCan.status, "cancelled")
        self.assertEqual(futureCan.status_text, "This event has been cancelled.")

    def testWhen(self):
        self.assertEqual(self.cancellation.when, "Wednesday 1st of February 1989 at 1pm to 3:30pm")

    def testAt(self):
        self.assertEqual(self.cancellation.at.strip(), "1pm")



# ------------------------------------------------------------------------------
# Joyous calendar models
# ------------------------------------------------------------------------------
import datetime as dt
import calendar
from django.conf import settings
from django.http import Http404
from django.shortcuts import render, redirect
from django.utils import timezone
from wagtail.core.models import Page
from wagtail.core.fields import RichTextField
from wagtail.admin.edit_handlers import FieldPanel, MultiFieldPanel
from wagtail.contrib.routable_page.models import RoutablePageMixin, route
from wagtail.search import index
from ..utils.weeks import week_info, gregorian_to_week_date
from ..utils.weeks import weekday_abbr, weekday_name
from . import getAllEventsByDay
from . import getAllEventsByWeek
from . import getAllUpcomingEvents
from . import getAllPastEvents

# ------------------------------------------------------------------------------
# Calendar
# ------------------------------------------------------------------------------
DatePictures = {"YYYY":  r"((?:19|20)\d\d)",
                "MM":    r"(1[012]|0?[1-9])",
                "Mmm":   r"({})".format("|".join(calendar.month_abbr[1:])),
                "DD":    r"(3[01]|[12]\d|0?[1-9])",
                "WW":    r"(5[0-3]|[1-4]\d|0?[1-9])"}

class CalendarPage(RoutablePageMixin, Page):
    subpage_types = ['joyous.SimpleEventPage',
                     'joyous.MultidayEventPage',
                     'joyous.RecurringEventPage']

    intro = RichTextField(blank=True)

    search_fields = Page.search_fields
    content_panels = Page.content_panels + [
        FieldPanel('intro', classname="full"),
        ]
    settings_panels = Page.settings_panels + [
        MultiFieldPanel([
            ], "Warning! this feature is experimental : Import"),
        MultiFieldPanel([
            ], "Export"),
        ]

    @route(r"^$")
    @route(r"^{YYYY}/$".format(**DatePictures))
    def routeDefault(self, request, year=None):
        eventsView = getattr(settings, "JOYOUS_DEFAULT_EVENTS_VIEW", "Monthly")
        if eventsView == "List":
            return self.serveUpcoming(request)
        elif eventsView == "Weekly":
            return self.serveWeek(request, year)
        else:
            return self.serveMonth(request, year)

    @route(r"(?i)^{YYYY}/{Mmm}/$".format(**DatePictures))
    def routeByMonthAbbr(self, request, year, monthAbbr):
        month = calendar.month_abbr[:].index(monthAbbr.title())
        return self.serveMonth(request, year, month)

    @route(r"^month/$")
    @route(r"^{YYYY}/{MM}/$".format(**DatePictures))
    def serveMonth(self, request, year=None, month=None):
        myurl = self.get_url(request)
        def myUrl(urlYear, urlMonth):
            return myurl + self.reverse_subpage('serveMonth',
                                                args=[urlYear, urlMonth])

        today = timezone.localdate()
        if year is None: year = today.year
        if month is None: month = today.month
        year = int(year)
        month = int(month)

        if year == today.year and month == today.month:
            weekNum = gregorian_to_week_date(today)[1]
        else:
            weekNum = gregorian_to_week_date(dt.date(year, month, 7))[1]
        weeklyUrl = myurl + self.reverse_subpage('serveWeek',
                                                 args=[year, weekNum])
        listUrl = myurl + self.reverse_subpage('serveUpcoming')

        prevMonth = month - 1
        prevMonthYear = year
        if prevMonth == 0:
            prevMonth = 12
            prevMonthYear -= 1

        nextMonth = month + 1
        nextMonthYear = year
        if nextMonth == 13:
            nextMonth = 1
            nextMonthYear += 1

        return render(request, "joyous/calendar_month.html",
                      {'self':         self,
                       'page':         self,
                       'year':         year,
                       'month':        month,
                       'today':        today,
                       'yesterday':    today - dt.timedelta(1),
                       'lastweek':     today - dt.timedelta(7),
                       'prevMonthUrl': myUrl(prevMonthYear, prevMonth),
                       'nextMonthUrl': myUrl(nextMonthYear, nextMonth),
                       'prevYearUrl':  myUrl(year - 1, month),
                       'nextYearUrl':  myUrl(year + 1, month),
                       'weeklyUrl':    weeklyUrl,
                       'listUrl':      listUrl,
                       'thisMonthUrl': myUrl(today.year, today.month),
                       'monthName':    calendar.month_name[month],
                       'weekdayAbbr':  weekday_abbr,
                       'events':       self._getEventsByWeek(year, month)})

    @route(r"^week/$")
    @route(r"^{YYYY}/W{WW}/$".format(**DatePictures))
    def serveWeek(self, request, year=None, week=None):
        myurl = self.get_url(request)
        def myUrl(urlYear, urlWeek):
            return myurl + self.reverse_subpage('serveWeek',
                                                args=[urlYear, urlWeek])

        today = timezone.localdate()
        thisYear, thisWeekNum, _ = gregorian_to_week_date(today)
        if year is None: year = thisYear
        if week is None: week = thisWeekNum
        year = int(year)
        week = int(week)

        firstDay, lastDay, prevYearNumWeeks, yearNumWeeks = week_info(year, week)
        eventsInWeek = self._getEventsByDay(firstDay, lastDay)
        monthlyUrl = myurl + self.reverse_subpage('serveMonth',
                                                  args=[firstDay.year, firstDay.month])
        listUrl = myurl + self.reverse_subpage('serveUpcoming')
        if week == 53 and yearNumWeeks == 52:
            year += 1
            week = 1

        prevWeek = week - 1
        prevWeekYear = year
        if prevWeek == 0:
            prevWeek = prevYearNumWeeks
            prevWeekYear -= 1

        nextWeek = week + 1
        nextWeekYear = year
        if nextWeek > yearNumWeeks:
            nextWeek = 1
            nextWeekYear += 1

        return render(request, "joyous/calendar_week.html",
                      {'self':         self,
                       'page':         self,
                       'year':         year,
                       'week':         week,
                       'today':        today,
                       'yesterday':    today - dt.timedelta(1),
                       'prevWeekUrl':  myUrl(prevWeekYear, prevWeek),
                       'nextWeekUrl':  myUrl(nextWeekYear, nextWeek),
                       'prevYearUrl':  myUrl(year - 1, week),
                       'nextYearUrl':  myUrl(year + 1, week),
                       'thisWeekUrl':  myUrl(thisYear, thisWeekNum),
                       'monthlyUrl':   monthlyUrl,
                       'listUrl':      listUrl,
                       'weekName':     "Week {}".format(week),
                       'weekdayAbbr':  weekday_abbr,
                       'events':       [eventsInWeek]})

    @route(r"^day/$")
    @route(r"^{YYYY}/{MM}/{DD}/$".format(**DatePictures))
    def serveDay(self, request, year=None, month=None, dom=None):
        myurl = self.get_url(request)
        today = timezone.localdate()
        if year is None: year = today.year
        if month is None: month = today.month
        if dom is None: dom = today.day
        year = int(year)
        month = int(month)
        dom = int(dom)
        day = dt.date(year, month, dom)

        eventsOnDay = self._getEventsOnDay(day)
        if len(eventsOnDay.all_events) == 1:
            event = eventsOnDay.all_events[0].page
            return redirect(event.get_url(request))

        monthlyUrl = myurl + self.reverse_subpage('serveMonth',
                                                  args=[year, month])
        weekNum = gregorian_to_week_date(today)[1]
        weeklyUrl = myurl + self.reverse_subpage('serveWeek',
                                                 args=[year, weekNum])
        listUrl = myurl + self.reverse_subpage('serveUpcoming')

        return render(request, "joyous/calendar_list_day.html",
                      {'self':         self,
                       'page':         self,
                       'year':         year,
                       'month':        month,
                       'dom':          dom,
                       'day':          day,
                       'monthlyUrl':   monthlyUrl,
                       'weeklyUrl':    weeklyUrl,
                       'listUrl':      listUrl,
                       'monthName':    calendar.month_name[month],
                       'weekdayName':  calendar.day_name[month],
                       'events':       eventsOnDay})

    @route(r"^upcoming/$")
    def serveUpcoming(self, request):
        myurl = self.get_url(request)
        today = timezone.localdate()
        monthlyUrl = myurl + self.reverse_subpage('serveMonth',
                                                  args=[today.year, today.month])
        weekNum = gregorian_to_week_date(today)[1]
        weeklyUrl = myurl + self.reverse_subpage('serveWeek',
                                                 args=[today.year, weekNum])
        listUrl = myurl + self.reverse_subpage('servePast')
        events = getAllUpcomingEvents()

        return render(request, "joyous/calendar_list_upcoming.html",
                      {'self':         self,
                       'page':         self,
                       'today':        today,
                       'weeklyUrl':    weeklyUrl,
                       'monthlyUrl':   monthlyUrl,
                       'listUrl':      listUrl,
                       'events':       events})

    @route(r"^past/$")
    def servePast(self, request):
        myurl = self.get_url(request)
        today = timezone.localdate()
        monthlyUrl = myurl + self.reverse_subpage('serveMonth',
                                                  args=[today.year, today.month])
        weekNum = gregorian_to_week_date(today)[1]
        weeklyUrl = myurl + self.reverse_subpage('serveWeek',
                                                 args=[today.year, weekNum])
        listUrl = myurl + self.reverse_subpage('serveUpcoming')
        events = getAllPastEvents()

        return render(request, "joyous/calendar_list_past.html",
                      {'self':         self,
                       'page':         self,
                       'today':        today,
                       'weeklyUrl':    weeklyUrl,
                       'monthlyUrl':   monthlyUrl,
                       'listUrl':      listUrl,
                       'events':       events})

    def _getEventsOnDay(self, day):
        return getAllEventsByDay(day, day)[0]

    def _getEventsByDay(self, firstDay, lastDay):
        return getAllEventsByDay(firstDay, lastDay)

    def _getEventsByWeek(self, year, month):
        return getAllEventsByWeek(year, month)

    @route(r"^mini/{YYYY}/{MM}/$".format(**DatePictures))
    def serveMiniMonth(self, request, year=None, month=None):
        if not request.is_ajax():
            raise Http404("/mini/ is for ajax requests only")

        today = timezone.localdate()
        if year is None: year = today.year
        if month is None: month = today.month
        year = int(year)
        month = int(month)

        prevMonth = month - 1
        prevMonthYear = year
        if prevMonth == 0:
            prevMonth = 12
            prevMonthYear -= 1

        nextMonth = month + 1
        nextMonthYear = year
        if nextMonth == 13:
            nextMonth = 1
            nextMonthYear += 1

        return render(request, "joyous/includes/minicalendar.html",
                      {'self':         self,
                       'page':         self,
                       'today':        today,
                       'year':         year,
                       'month':        month,
                       'calendarUrl':  self.get_url(request),
                       'monthName':    calendar.month_name[month],
                       'weekdayInfo':  zip(weekday_abbr, weekday_name),
                       'events':       self._getEventsByWeek(year, month)})

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-02-28 01:49
from __future__ import unicode_literals

from django.db import migrations
import wagtail.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('joyous', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='simpleeventpage',
            options={'verbose_name': 'Event Page'},
        ),
        migrations.AlterField(
            model_name='grouppage',
            name='content',
            field=wagtail.core.fields.RichTextField(blank=True, default='', help_text='An area of text for whatever you like'),
        ),
    ]



# ------------------------------------------------------------------------------
# Joyous events models
# ------------------------------------------------------------------------------
import datetime as dt
import calendar
from collections import namedtuple
from itertools import groupby
from operator import attrgetter
from django.conf import settings
from django.db import models
from django.http import HttpResponse
from django.utils.html import format_html
from django.utils import timezone
from wagtail.core.models import Page
from wagtail.core.fields import RichTextField
from wagtail.admin.edit_handlers import (FieldPanel, MultiFieldPanel,
        PageChooserPanel)
from wagtail.images.edit_handlers import ImageChooserPanel
from wagtail.images import get_image_model_string
from wagtail.search import index
from wagtail.admin.forms import WagtailAdminPageForm
from ..holidays.parser import parseHolidays
from ..utils.telltime import (getAwareDatetime, getLocalDatetime,
        getLocalDateAndTime, getLocalDate, getLocalTime)
from ..utils.telltime import timeFrom, timeTo
from ..utils.telltime import timeFormat, dateFormat
from ..utils.weeks import week_of_month
# from ..utils.ical import export_event
from ..recurrence import RecurrenceField
from ..recurrence import ExceptionDatePanel
from ..widgets import TimeInput
from .groups import get_group_model_string, get_group_model

# using django-timezone-field
from timezone_field import TimeZoneField
# TODO consider using django-timezone-utils?
# from timezone_utils.fields import TimeZoneField

try:
    # Use wagtailgmaps for location if it is installed
    # but don't depend upon it
    settings.INSTALLED_APPS.index('wagtailgmaps')
    from wagtailgmaps.edit_handlers import MapFieldPanel
except (ValueError, ImportError):
    MapFieldPanel = FieldPanel

# ------------------------------------------------------------------------------
ThisEvent = namedtuple("ThisEvent", "title page")

class EventsOnDay(namedtuple("EODBase", "date days_events continuing_events")):
    holidays = parseHolidays(getattr(settings, "JOYOUS_HOLIDAYS", ""))

    @property
    def all_events(self):
        return self.days_events + self.continuing_events

    @property
    def preview(self):
        return ", ".join(event.title for event in self.all_events)[:100]

    @property
    def weekday(self):
        return calendar.day_abbr[self.date.weekday()].lower()

    @property
    def holiday(self):
        return self.holidays.get(self.date)

# ------------------------------------------------------------------------------
_1day = dt.timedelta(days=1)

def _getEventsByDay(date_from, eventsByDaySrcs):
    eventsByDay = []
    day = date_from
    for srcs in zip(*eventsByDaySrcs):
        days_events       = []
        continuing_events = []
        for src in srcs:
            days_events += src.days_events
            continuing_events += src.continuing_events
        def sortByTime(thisEvent):
            time_from = thisEvent.page.time_from
            if time_from is None:
                time_from = dt.time.max
            return time_from
        days_events.sort(key=sortByTime)
        eventsByDay.append(EventsOnDay(day, days_events, continuing_events))
        day += _1day
    return eventsByDay

def getAllEventsByDay(date_from, date_to):
    simpleEvents    = SimpleEventPage.getEventsByDay(date_from, date_to)
    multidayEvents  = MultidayEventPage.getEventsByDay(date_from, date_to)
    recurringEvents = RecurringEventPage.getEventsByDay(date_from, date_to)
    postponedEvents = PostponementPage.getEventsByDay(date_from, date_to)
    allEvents = _getEventsByDay(date_from, (simpleEvents, multidayEvents,
                                            recurringEvents, postponedEvents))
    return allEvents

def _getEventsByWeek(year, month, eventsByDaySrc):
    weeks = []
    firstDay = dt.date(year, month, 1)
    lastDay  = dt.date(year, month, calendar.monthrange(year, month)[1])
    def calcWeekOfMonth(evod):
        return week_of_month(evod.date)
    events = eventsByDaySrc(firstDay, lastDay)
    for weekOfMonth, group in groupby(events, calcWeekOfMonth):
        week = list(group)
        if len(week) < 7:
            padding = [None] * (7 - len(week))
            if weekOfMonth == 0:
                week = padding + week
            else:
                week += padding
        weeks.append(week)
    return weeks

def getAllEventsByWeek(year, month):
    return _getEventsByWeek(year, month, getAllEventsByDay)

# ------------------------------------------------------------------------------
def _getUpcomingEvents(simpleEventsQry=None,
                       multidayEventsQry=None,
                       recurringEventsQry=None,
                       postponedEventsQry=None,
                       extraInfoQry=None):
    todaySomewhere = dt.datetime.utcnow().date() - _1day
    events = []
    if simpleEventsQry is not None:
        for event in simpleEventsQry.live().filter(date__gte=todaySomewhere):
            if event._upcoming_datetime_from:
                events.append(ThisEvent(event.title, event))
    if multidayEventsQry is not None:
        for event in multidayEventsQry.live().filter(date_from__gte=todaySomewhere):
            if event._upcoming_datetime_from:
                events.append(ThisEvent(event.title, event))
    if postponedEventsQry is not None:
        for event in postponedEventsQry.live().filter(date__gte=todaySomewhere):
            if event._upcoming_datetime_from:
                events.append(ThisEvent(event.postponement_title, event))
    if extraInfoQry is not None:
        for event in extraInfoQry.live().filter(except_date__gte=todaySomewhere)     \
                                        .exclude(extra_title=""):
            if event._upcoming_datetime_from:
                events.append(ThisEvent(event.extra_title, event))
    if recurringEventsQry is not None:
        for event in recurringEventsQry.live():
            if event._upcoming_datetime_from:
                events.append(ThisEvent(event.title, event))
    return events

def getAllUpcomingEvents(home=None):
    if home is None:
        events =  _getUpcomingEvents(SimpleEventPage.objects,
                                     MultidayEventPage.objects,
                                     RecurringEventPage.objects,
                                     PostponementPage.objects,
                                     ExtraInfoPage.objects)
    else:
        events = _getUpcomingEvents(SimpleEventPage.objects.descendant_of(home),
                                    MultidayEventPage.objects.descendant_of(home),
                                    RecurringEventPage.objects.descendant_of(home),
                                    PostponementPage.objects.descendant_of(home),
                                    ExtraInfoPage.objects.descendant_of(home))
    events.sort(key=attrgetter('page._upcoming_datetime_from'))
    return events

def getGroupUpcomingEvents(group):
    # Get events that are a child of a group page, or a postponement or extra
    # info a child of the recurring event child of the group (using descendant_of)
    events = _getUpcomingEvents(SimpleEventPage.objects.child_of(group),
                                MultidayEventPage.objects.child_of(group),
                                RecurringEventPage.objects.child_of(group),
                                PostponementPage.objects.descendant_of(group),
                                ExtraInfoPage.objects.descendant_of(group))

    # Get events that are linked to a group page, or a postponement or extra
    # info a child of the recurring event linked to a group (the long way)
    events += _getUpcomingEvents(group.simpleeventpage_events,
                                 group.multidayeventpage_events)
    rrEvents = _getUpcomingEvents(recurringEventsQry=group.recurringeventpage_events)
    events += rrEvents
    for rrEvent in rrEvents:
        postponedEventsQry = PostponementPage.objects.child_of(rrEvent.page)
        extraInfoQry       = ExtraInfoPage.objects.child_of(rrEvent.page)
        events += _getUpcomingEvents(postponedEventsQry=postponedEventsQry,
                                     extraInfoQry=extraInfoQry)
    events.sort(key=attrgetter('page._upcoming_datetime_from'))
    return events

def _getPastEvents(simpleEventsQry=None,
                   multidayEventsQry=None,
                   recurringEventsQry=None,
                   postponedEventsQry=None,
                   extraInfoQry=None):
    todaySomewhere = dt.datetime.utcnow().date() + _1day
    events = []
    if simpleEventsQry is not None:
        for event in simpleEventsQry.live().filter(date__lte=todaySomewhere):
            if event._past_datetime_from:
                events.append(ThisEvent(event.title, event))
    if multidayEventsQry is not None:
        for event in multidayEventsQry.live().filter(date_from__lte=todaySomewhere):
            if event._past_datetime_from:
                events.append(ThisEvent(event.title, event))
    if postponedEventsQry is not None:
        for event in postponedEventsQry.live().filter(date__lte=todaySomewhere):
            if event._past_datetime_from:
                events.append(ThisEvent(event.postponement_title, event))
    if extraInfoQry is not None:
        for event in extraInfoQry.live().filter(except_date__lte=todaySomewhere)    \
                                        .exclude(extra_title=""):
            if event._past_datetime_from:
                events.append(ThisEvent(event.extra_title, event))
    if recurringEventsQry is not None:
        for event in recurringEventsQry.live():
            if event._past_datetime_from:
                events.append(ThisEvent(event.title, event))
    return events

def getAllPastEvents():
    events = _getPastEvents(SimpleEventPage.objects,
                            MultidayEventPage.objects,
                            RecurringEventPage.objects,
                            PostponementPage.objects,
                            ExtraInfoPage.objects)
    events.sort(key=attrgetter('page._past_datetime_from'), reverse=True)
    return events

# ------------------------------------------------------------------------------
class EventCategory(models.Model):
    class Meta:
        ordering = ["name"]
        verbose_name = "Event Category"
        verbose_name_plural = "Event Categories"

    code = models.CharField(max_length=4, unique=True)
    name = models.CharField(max_length=80)

    def __str__(self):
        return self.name

# ------------------------------------------------------------------------------
class EventPageForm(WagtailAdminPageForm):
    def clean(self):
        cleaned_data = super().clean()
        self._checkStartBeforeEnd(cleaned_data)
        return cleaned_data

    def _checkStartBeforeEnd(self, cleaned_data):
        startTime = timeFrom(cleaned_data.get('time_from'))
        endTime   = timeTo(cleaned_data.get('time_to'))
        if startTime > endTime:
            self.add_error('time_to', "Event cannot end before it starts")

def _get_default_timezone():
    return timezone.get_default_timezone()

class EventBase(models.Model):
    class Meta:
        abstract = True

    category = models.ForeignKey(EventCategory,
                                 related_name="+",
                                 verbose_name="Category",
                                 on_delete=models.SET_NULL,
                                 blank=True, null=True)
    image = models.ForeignKey(get_image_model_string(),
                              null=True, blank=True,
                              on_delete=models.SET_NULL,
                              related_name='+')

    time_from = models.TimeField("Start time", null=True, blank=True)
    time_to = models.TimeField("End time", null=True, blank=True)
    # No you can't set different timezones for time_from and time_to
    tz = TimeZoneField(verbose_name="Time zone",
                       default=_get_default_timezone)

    group_page  = models.ForeignKey(get_group_model_string(),
                                    null=True, blank=True,
                                    on_delete=models.SET_NULL,
                                    related_name="%(class)s_events",
                                    related_query_name="%(class)s_event")
    details  = RichTextField(blank=True)
    location = models.CharField(max_length=255, blank=True)
    website = models.URLField(blank=True)

    search_fields = Page.search_fields + [
        index.SearchField('location'),
        index.SearchField('details'),
    ]

    content_panels1 = [
        FieldPanel('details', classname="full"),
        MapFieldPanel('location'),
        FieldPanel('website'),
    ]
    if getattr(settings, "JOYOUS_GROUP_SELECTABLE", False):
        content_panels1.append(PageChooserPanel('group_page'))

    # adding the event as a child of a group automatically assigns the event to
    # that group.
    @property
    def group(self):
        retval = None
        parent = self.get_parent()
        Group = get_group_model()
        if issubclass(parent.specific_class, Group):
            retval = parent.specific
        if retval is None:
            retval = self.group_page
        return retval

    @property
    def _upcoming_datetime_from(self):
        fromDt = self._getFromDt()
        return fromDt if fromDt >= timezone.localtime() else None

    @property
    def _past_datetime_from(self):
        fromDt = self._getFromDt()
        return fromDt if fromDt < timezone.localtime() else None

    @property
    def status(self):
        raise NotImplementedError()

    @property
    def status_text(self):
        status = self.status
        if status == "finished":
            return "This event has finished."
        elif status == "started":
            return "This event has started."
        else:
            return ""

    @classmethod
    def _removeContentPanels(cls, remove):
        if type(remove) is str:
            remove = [remove]
        cls.content_panels = [panel for panel in cls.content_panels
                              if getattr(panel, "field_name", None) not in remove]

    def _getLocalWhen(self, date_from, date_to=None):
        dateFrom, timeFrom = getLocalDateAndTime(date_from, self.time_from,
                                                 self.tz, dt.time.min)

        if date_to is not None:
            dateTo, timeTo = getLocalDateAndTime(date_to, self.time_to, self.tz)
        else:
            if self.time_to is not None:
                dateTo, timeTo = getLocalDateAndTime(date_from, self.time_to, self.tz)
            else:
                dateTo = dateFrom
                timeTo = None

        if dateFrom == dateTo:
            retval = "{} {}".format(dateFormat(dateFrom),
                                    timeFormat(timeFrom, timeTo, "at "))
        else:
            retval = "{} {}".format(dateFormat(dateFrom),
                                    timeFormat(timeFrom, prefix="at "))
            retval = "{} to {} {}".format(retval.strip(),
                                          dateFormat(dateTo),
                                          timeFormat(timeTo, prefix="at "))
        return retval.strip()

    def _getFromDt(self):
        raise NotImplementedError()

def removeContentPanels(remove):
    SimpleEventPage._removeContentPanels(remove)
    MultidayEventPage._removeContentPanels(remove)
    RecurringEventPage._removeContentPanels(remove)
    PostponementPage._removeContentPanels(remove)

# ------------------------------------------------------------------------------
class SimpleEventPage(Page, EventBase):
    class Meta:
        verbose_name = "Event Page"

    parent_page_types = ["joyous.CalendarPage",
                         get_group_model_string()]
    subpage_types = []
    base_form_class = EventPageForm

    date    = models.DateField("Date", default=dt.date.today)

    content_panels = Page.content_panels + [
        FieldPanel('category'),
        ImageChooserPanel('image'),
        FieldPanel('date'),
        FieldPanel('time_from', widget=TimeInput),
        FieldPanel('time_to', widget=TimeInput),
        FieldPanel('tz'),
        ] + EventBase.content_panels1

    @classmethod
    def getEventsByDay(cls, date_from, date_to):
        ordFrom =  date_from.toordinal()
        ordTo   =  date_to.toordinal()
        # TODO think about changing to the same algorithm as for multiday event
        # This works and is arguably faster [O(#pages) < O(#days * #pages) ???]
        # but the other one is easier to read
        events = [EventsOnDay(dt.date.fromordinal(ord), [], [])
                  for ord in range(ordFrom, ordTo+1)]

        dateRange = (date_from-_1day, date_to+_1day)
        pages = SimpleEventPage.objects.live().filter(date__range=dateRange)
        for page in pages:
            # convert event dates to local dates
            fromDate = getLocalDate(page.date, page.time_from, page.tz)
            toDate   = getLocalDate(page.date, page.time_to, page.tz)
            dayNum = fromDate.toordinal() - ordFrom
            thisEvent = ThisEvent(page.title, page)
            if 0 <= dayNum <= ordTo - ordFrom:
                events[dayNum].days_events.append(thisEvent)
            if fromDate != toDate:
                if 0 <= dayNum+1 <= ordTo - ordFrom:
                    events[dayNum+1].continuing_events.append(thisEvent)
        return events

    @property
    def status(self):
        myNow = timezone.localtime(timezone=self.tz)
        if getAwareDatetime(self.date, self.time_to, self.tz) < myNow:
            return "finished"
        elif getAwareDatetime(self.date, self.time_from, self.tz) < myNow:
            return "started"
        return None

    @property
    def when(self):
        return self._getLocalWhen(self.date)

    @property
    def at(self):
        timeFrom = getLocalTime(self.date, self.time_from, self.tz)
        return timeFormat(timeFrom)

    def _getFromDt(self):
        return getLocalDatetime(self.date, self.time_from, self.tz)

# ------------------------------------------------------------------------------
# TODO I *could* replace SimpleEventPage with MultidayEventPage

class MultidayEventPageForm(EventPageForm):
    def _checkStartBeforeEnd(self, cleaned_data):
        startDate = cleaned_data.get('date_from', dt.date.min)
        endDate   = cleaned_data.get('date_to', dt.date.max)
        if startDate > endDate:
            self.add_error('date_to', "Event cannot end before it starts")
        elif startDate == endDate:
            super()._checkStartBeforeEnd(cleaned_data)

class MultidayEventPage(Page, EventBase):
    class Meta:
        verbose_name = "Multiday Event Page"

    parent_page_types = ["joyous.CalendarPage",
                         get_group_model_string()]
    subpage_types = []
    base_form_class = MultidayEventPageForm

    date_from = models.DateField("Start date", default=dt.date.today)
    date_to = models.DateField("End date", default=dt.date.today)

    content_panels = Page.content_panels + [
        FieldPanel('category'),
        ImageChooserPanel('image'),
        FieldPanel('date_from'),
        FieldPanel('time_from', widget=TimeInput),
        FieldPanel('date_to'),
        FieldPanel('time_to', widget=TimeInput),
        FieldPanel('tz'),
        ] + EventBase.content_panels1

    @classmethod
    def getEventsByDay(cls, date_from, date_to):
        events = []
        ordFrom =  date_from.toordinal()
        ordTo   =  date_to.toordinal()
        # TODO think about using this algorithm for the other event classes
        days = [dt.date.fromordinal(ord) for ord in range(ordFrom, ordTo+1)]
        pages = MultidayEventPage.objects.live()                             \
                                 .filter(date_to__gte   = date_from-_1day)   \
                                 .filter(date_from__lte = date_to+_1day)
        for day in days:
            days_events = []
            continuing_events = []
            for page in pages:
                # convert event dates to local dates
                fromDate = getLocalDate(page.date_from, page.time_from, page.tz)
                toDate   = getLocalDate(page.date_to, page.time_to, page.tz)
                if fromDate == day:
                    days_events.append(ThisEvent(page.title, page))
                elif fromDate < day <= toDate:
                    continuing_events.append(ThisEvent(page.title, page))
            events.append(EventsOnDay(day, days_events, continuing_events))
        return events

    @property
    def status(self):
        myNow = timezone.localtime(timezone=self.tz)
        if getAwareDatetime(self.date_to, self.time_to, self.tz) < myNow:
            return "finished"
        elif getAwareDatetime(self.date_from, self.time_from, self.tz) < myNow:
            return "started"
        return None

    @property
    def when(self):
        return self._getLocalWhen(self.date_from, self.date_to)

    @property
    def at(self):
        timeFrom = getLocalTime(self.date_from, self.time_from, self.tz)
        return timeFormat(timeFrom)

    def _getFromDt(self):
        return getLocalDatetime(self.date_from, self.time_from, self.tz)

# ------------------------------------------------------------------------------
class RecurringEventPage(Page, EventBase):
    class Meta:
        verbose_name = "Recurring Event Page"

    parent_page_types = ["joyous.CalendarPage",
                         get_group_model_string()]
    subpage_types = ['joyous.ExtraInfoPage',
                     'joyous.CancellationPage',
                     'joyous.PostponementPage']
    base_form_class = EventPageForm

    # FIXME So that Fred can't cancel Barney's event
    # owner_subpages_only = True

    repeat  = RecurrenceField()

    # TODO 
    # exclude_holidays = models.BooleanField(default=False)
    # exclude_holidays.help_text = "Cancel any occurence of this event on a public holiday"

    content_panels = Page.content_panels + [
        FieldPanel('category'),
        ImageChooserPanel('image'),
        FieldPanel('repeat'),
        FieldPanel('time_from', widget=TimeInput),
        FieldPanel('time_to', widget=TimeInput),
        FieldPanel('tz'),
        ] + EventBase.content_panels1

    @property
    def next_date(self):
        """
        Date when this event is next scheduled to occur
        (Does not include postponements, but does exclude cancellations)
        """
        # TODO default to max or min ?!?!?!!!
        nextDt = self.__localAfter(timezone.localtime(), dt.time.min)
        if nextDt is not None:
            return nextDt.date()
        else:
            return None

    @property
    def _upcoming_datetime_from(self):
        nextDt = self.__localAfter(timezone.localtime(), dt.time.max,
                                   excludeCancellations=True,
                                   excludeExtraInfo=True)
        return nextDt

    @property
    def prev_date(self):
        """
        Date when this event last occurred
        (Does not include postponements, but does exclude cancellations)
        """
        # TODO default to max or min ?!?!?!!!
        prevDt = self.__localBefore(timezone.localtime(), dt.time.min)
        if prevDt is not None:
            return prevDt.date()
        else:
            return None

    @property
    def _past_datetime_from(self):
        prevDt = self.__localBefore(timezone.localtime(), dt.time.max,
                                    excludeCancellations=True,
                                    excludeExtraInfo=True)
        return prevDt

    @property
    def next_on(self):
        """
        Formatted date/time of when this event (including any postponements)
        will next be on
        """
        retval = None
        # TODO default to max or min ?!?!?!!!
        nextDt, event = self.__localAfterOrPostponedTo(timezone.localtime(),
                                                       dt.time.min)
        if nextDt is not None:
            timeFrom = nextDt.time() if event.time_from is not None else None
            retval = "{} {}".format(dateFormat(nextDt.date()),
                                    timeFormat(timeFrom, prefix="at "))
            if event is not self:
                retval = format_html('<a class="inline-link" href="{}">{}</a>',
                                     event.url, retval)
        return retval

    @property
    def status(self):
        myNow = timezone.localtime(timezone=self.tz)
        if self.repeat.until:
            untilDt = getAwareDatetime(self.repeat.until, self.time_to, self.tz)
            if untilDt < myNow:
                return "finished"
        todayStart = getAwareDatetime(myNow.date(), dt.time.min, self.tz)
        eventStart, event = self.__afterOrPostponedTo(todayStart)
        if eventStart is None:
            # the last occurences must have been cancelled
            return "finished"
        eventFinish = getAwareDatetime(eventStart.date(), event.time_to, self.tz)
        if (event.time_from is not None and
            eventStart < myNow < eventFinish):
            # if there are two occurences on the same day then we may miss
            # that one of them has started
            return "started"
        if (self.repeat.until and eventFinish < myNow and
            self.__afterOrPostponedTo(myNow)[0] is None):
            # only just wound up, the last occurence was earlier today
            return "finished"
        return None

    @property
    def status_text(self):
        status = self.status
        if status == "finished":
            return "These events have finished."
        else:
            return super().status_text

    @property
    def when(self):
        # TODO: is calculating the next date too expensive?????
        # if so change it to just be based on today's date
        # fromTime = getLocalTime(timezone.localdate(), self.time_from)
        # toTime = getLocalTime(timezone.localdate(), self.time_to)
        offset   = 0
        timeFrom = None
        timeTo   = None
        fromDt   = self.__after(timezone.localtime(timezone=self.tz))
        if fromDt is not None:
            offset = timezone.localtime(fromDt).toordinal() - fromDt.toordinal()
            if self.time_from is not None:
                timeFrom = getLocalTime(fromDt.date(), self.time_from, self.tz)
            if self.time_to is not None:
                timeTo = getLocalTime(fromDt.date(), self.time_to, self.tz)
        retval = "{} {}".format(self.repeat._getWhen(offset),
                                timeFormat(timeFrom, timeTo, "at "))
        return retval.strip()

    @property
    def at(self):
        # TODO as above
        timeFrom = None
        fromDt   = self.__after(timezone.localtime(timezone=self.tz))
        if fromDt is not None:
            timeFrom = getLocalTime(fromDt.date(), self.time_from, self.tz)
        return timeFormat(timeFrom)

    @property
    def future_exceptions(self):
        """
        Returns all future extra info, cancellations and postponements created
        for this recurring event
        """
        retval = []
        # We know all future exception dates are in the parent time zone
        myToday = timezone.localdate(timezone=self.tz)

        for extraInfo in ExtraInfoPage.objects.live().child_of(self)         \
                                      .filter(except_date__gte=myToday):
            retval.append(ThisEvent(extraInfo.extra_title, extraInfo))
        for cancellation in CancellationPage.objects.live().child_of(self)   \
                                            .filter(except_date__gte=myToday):
            postponement = getattr(cancellation, "postponementpage", None)
            if postponement:
                retval.append(ThisEvent(postponement.postponement_title, postponement))
            else:
                retval.append(ThisEvent(cancellation.cancellation_title, cancellation))
        retval.sort(key=attrgetter('page.except_date'))
        return retval

    # TODO add def _localOccursOn(self, localDate)?

    def _occursOn(self, myDate):
        """
        Returns true iff this event occurs on this date (in event's own timezone)
        (Does not include postponements, but does exclude cancellations)
        """
        # TODO analyse which is faster (rrule or db) and test that first
        if myDate not in self.repeat:
            return False
        if CancellationPage.objects.live().child_of(self)                    \
                           .filter(except_date=myDate).exists():
            return False
        return True

    def _getMyFirstDatetimeFrom(self):
        myStartDt = getAwareDatetime(self.repeat.dtstart, None,
                                     self.tz, dt.time.min)
        return self.__after(myStartDt)

    def _getMyFirstDatetimeTo(self):
        myFirstDt = self._getMyFirstDatetimeFrom()
        if myFirstDt is not None:
            return getAwareDatetime(myFirstDt.date(), self.time_to,
                                    self.tz, dt.time.max)
        else:
            return None

    @classmethod
    def getEventsByDay(cls, date_from, date_to):
        ordFrom =  date_from.toordinal()
        ordTo   =  date_to.toordinal()
        # TODO As with simple event, think about changing to the same algorithm
        # as used for multiday event
        events = [EventsOnDay(dt.date.fromordinal(ord), [], [])
                  for ord in range(ordFrom, ordTo+1)]
        pages = RecurringEventPage.objects.live()
        for page in pages:
            exceptions = page.__getExceptions(date_from-_1day, date_to+_1day)

            for occurence in page.repeat.between(date_from-_1day,
                                                 date_to+_1day, True):
                thisEvent = None
                exception = exceptions.get(occurence)
                if exception:
                    if exception.title:
                        thisEvent = exception
                else:
                    thisEvent = ThisEvent(page.title, page)

                if thisEvent:
                    fromDate = getLocalDate(occurence, page.time_from, page.tz)
                    todDate  = getLocalDate(occurence, page.time_to, page.tz)
                    dayNum = fromDate.toordinal() - ordFrom
                    if 0 <= dayNum <= ordTo - ordFrom:
                        events[dayNum].days_events.append(thisEvent)
                    if fromDate != todDate:
                        if 0 <= dayNum+1 <= ordTo - ordFrom:
                            events[dayNum+1].continuing_events.append(thisEvent)
        return events

    def __getExceptions(self, date_from, date_to):
        exceptions = {}

        for extraInfo in ExtraInfoPage.objects.live().child_of(self)         \
                         .filter(except_date__range=(date_from, date_to)):
            title = extraInfo.extra_title or self.title
            exceptions[extraInfo.except_date] = ThisEvent(title, extraInfo)

        for cancellation in CancellationPage.objects.live().child_of(self)   \
                         .filter(except_date__range=(date_from, date_to)):
            title = cancellation.cancellation_title
            exceptions[cancellation.except_date] = ThisEvent(title, cancellation)

        return exceptions

    def __localAfterOrPostponedTo(self, fromDt, timeDefault=dt.time.min):
        myFromDt, event = self.__afterOrPostponedTo(fromDt.astimezone(self.tz))
        if myFromDt is not None:
            localFromDt = getLocalDatetime(myFromDt.date(), event.time_from,
                                           self.tz, timeDefault)
            return (localFromDt, event)
        else:
            return (None, event)

    def __afterOrPostponedTo(self, fromDt):
        after = self.__after(fromDt)
        # We know all postponement exception dates are in the parent time zone
        if after:
            # is there a postponed event before that?
            # nb: range is inclusive
            postponements = PostponementPage.objects.live().child_of(self)                \
                                     .filter(date__range=(fromDt.date(), after.date()))   \
                                     .order_by('date', 'time_from')
            for postponement in postponements:
                postDt = getAwareDatetime(postponement.date,
                                          postponement.time_from,
                                          self.tz, dt.time.min)
                postDtMax = getAwareDatetime(postponement.date,
                                             postponement.time_from,
                                             self.tz, dt.time.max)
                if postDt < after and postDtMax >= fromDt:
                    return (postDt, postponement)
        else:
            # is there a postponed event then?
            postponements = PostponementPage.objects.live().child_of(self)   \
                                     .filter(date__gte=fromDt.date())        \
                                     .order_by('date', 'time_from')
            for postponement in postponements:
                postDt = getAwareDatetime(postponement.date,
                                          postponement.time_from,
                                          self.tz, dt.time.min)
                postDtMax = getAwareDatetime(postponement.date,
                                             postponement.time_from,
                                             self.tz, dt.time.max)
                if postDtMax >= fromDt:
                    return (postDt, postponement)

        if after is not None:
            return (after, self)
        else:
            return (None, None)

    def __localAfter(self, fromDt, timeDefault=dt.time.min, **kwargs):
        myFromDt = self.__after(fromDt.astimezone(self.tz), **kwargs)
        if myFromDt is not None:
            return getLocalDatetime(myFromDt.date(), self.time_from,
                                    self.tz, timeDefault)
        else:
            return None

    def __after(self, fromDt, excludeCancellations=True, excludeExtraInfo=False):
        fromDate = fromDt.date()
        if self.time_from and self.time_from < fromDt.time():
            fromDate += dt.timedelta(days=1)
        exceptions = set()
        if excludeCancellations:
            for cancelled in CancellationPage.objects.live().child_of(self)  \
                                     .filter(except_date__gte=fromDate):
                exceptions.add(cancelled.except_date)
        if excludeExtraInfo:
            for info in ExtraInfoPage.objects.live().child_of(self)          \
                                     .filter(except_date__gte=fromDate)      \
                                     .exclude(extra_title=""):
                exceptions.add(info.except_date)
        for occurence in self.repeat.xafter(fromDate, inc=True):
            if occurence not in exceptions:
                # TODO or return a plain date?
                return getAwareDatetime(occurence, self.time_from,
                                        self.tz, dt.time.min)
        return None

    def __localBefore(self, fromDt, timeDefault=dt.time.min, **kwargs):
        myFromDt = self.__before(fromDt.astimezone(self.tz), **kwargs)
        if myFromDt is not None:
            return getLocalDatetime(myFromDt.date(), self.time_from,
                                    self.tz, timeDefault)
        else:
            return None

    def __before(self, fromDt, excludeCancellations=True, excludeExtraInfo=False):
        fromDate = fromDt.date()
        if self.time_from and self.time_from > fromDt.time():
            fromDate -= dt.timedelta(days=1)
        exceptions = set()
        if excludeCancellations:
            for cancelled in CancellationPage.objects.live().child_of(self)  \
                                     .filter(except_date__lte=fromDate):
                exceptions.add(cancelled.except_date)
        if excludeExtraInfo:
            for info in ExtraInfoPage.objects.live().child_of(self)          \
                                     .filter(except_date__lte=fromDate)      \
                                     .exclude(extra_title=""):
                exceptions.add(info.except_date)
        last = None
        for occurence in self.repeat:
            if occurence > fromDate:
                break
            if occurence not in exceptions:
                last = occurence

        if last is not None:
            return getAwareDatetime(last, self.time_from, self.tz, dt.time.min)
        else:
            return None

# TODO
# class MultidayReccuringEventPage(RecurringEventPage):

# ------------------------------------------------------------------------------
class EventExceptionPageForm(WagtailAdminPageForm):
    def _checkSlugAvailable(self, cleaned_data, slugName=None):
        if slugName is None:
            slugName = self.slugName
        description = getattr(self, 'description', "a {}".format(slugName))
        exceptDate = cleaned_data.get('except_date', "invalid")
        slug = "{}-{}".format(exceptDate, slugName)
        if not Page._slug_is_available(slug, self.parent_page, self.instance):
            self.add_error('except_date',
                           'That date already has {}'.format(description))

    def save(self, commit=True):
        name = getattr(self, 'name', self.slugName.title())
        page = super().save(commit=False)
        page.title = "{} for {}".format(name, dateFormat(page.except_date))
        page.slug = "{}-{}".format(page.except_date, self.slugName)
        if commit:
            page.save()
        return page

class EventExceptionBase(models.Model):
    class Meta:
        abstract = True

    # overrides is also the parent, but parent is not set until the
    # child is saved and added.  (NB: is published version of parent)
    overrides = models.ForeignKey('joyous.RecurringEventPage',
                                  null=True, blank=False,
                                  # can't set to CASCADE, so go with SET_NULL
                                  on_delete=models.SET_NULL,
                                  related_name='+')
    overrides.help_text = "The recurring event that we are updating."
    except_date = models.DateField('For Date')
    except_date.help_text = "For this date"

    # Original properties
    time_from   = property(attrgetter("overrides.time_from"))
    time_to     = property(attrgetter("overrides.time_to"))
    tz          = property(attrgetter("overrides.tz"))
    group       = property(attrgetter("overrides.group"))

    @property
    def overrides_repeat(self):
        return getattr(self.overrides, 'repeat', None)

    @property
    def localTitle(self):
        # TODO localize for language too?
        name = self.title.partition(" for ")[0]
        exceptDate = getLocalDate(self.except_date, self.time_from, self.tz)
        title = "{} for {}".format(name, dateFormat(exceptDate))
        return title

    @property
    def when(self):
        return EventBase._getLocalWhen(self, self.except_date)

    @property
    def at(self):
        timeFrom = getLocalTime(self.except_date, self.time_from, self.tz)
        return timeFormat(timeFrom)

# ------------------------------------------------------------------------------
class ExtraInfoPageForm(EventExceptionPageForm):
    name        = "Extra Information"
    description = name.lower()
    slugName    = "extra-info"

    def clean(self):
        cleaned_data = super().clean()
        self._checkSlugAvailable(cleaned_data)
        return cleaned_data

class ExtraInfoPage(Page, EventExceptionBase):
    class Meta:
        verbose_name = "Extra Event Information"

    parent_page_types = ["joyous.RecurringEventPage"]
    subpage_types = []
    base_form_class = ExtraInfoPageForm

    extra_title = models.CharField('Title', max_length=255, blank=True)
    extra_title.help_text = "A more specific title for this occurence (optional)"
    extra_information = RichTextField(blank=False)
    extra_information.help_text = "Information just for this date"

    # Note title is not displayed
    content_panels = [
        PageChooserPanel('overrides'),
        ExceptionDatePanel('except_date'),
        FieldPanel('extra_title', classname="full title"),
        FieldPanel('extra_information', classname="full"),
        ]
    promote_panels = []

    # Original properties
    category    = property(attrgetter("overrides.category"))
    image       = property(attrgetter("overrides.image"))
    location    = property(attrgetter("overrides.location"))
    website     = property(attrgetter("overrides.website"))

    @property
    def status(self):
        myNow = timezone.localtime(timezone=self.tz)
        if getAwareDatetime(self.except_date, self.time_to, self.tz) < myNow:
            return "finished"
        elif getAwareDatetime(self.except_date, self.time_from, self.tz) < myNow:
            return "started"
        return None

    @property
    def status_text(self):
        return EventBase.status_text.fget(self)

    @property
    def _upcoming_datetime_from(self):
        return self._checkFromDt(lambda fromDt:fromDt >= timezone.localtime())

    @property
    def _past_datetime_from(self):
        return self._checkFromDt(lambda fromDt:fromDt < timezone.localtime())

    def _checkFromDt(self, predicate):
        if not self.overrides._occursOn(self.except_date):
            return None
        fromDt = getLocalDatetime(self.except_date, self.time_from, self.tz)
        return fromDt if predicate(fromDt) else None

# ------------------------------------------------------------------------------
class CancellationPageForm(EventExceptionPageForm):
    slugName = "cancellation"

    def clean(self):
        cleaned_data = super().clean()
        self._checkSlugAvailable(cleaned_data)
        self._checkSlugAvailable(cleaned_data, "postponement")
        return cleaned_data

class CancellationPage(Page, EventExceptionBase):
    class Meta:
        verbose_name = "Cancellation"

    parent_page_types = ["joyous.RecurringEventPage"]
    subpage_types = []
    base_form_class = CancellationPageForm

    cancellation_title = models.CharField('Title', max_length=255, blank=True)
    cancellation_title.help_text = "Show in place of cancelled event "\
                                   "(Leave empty to show nothing)"
    cancellation_details = RichTextField('Details', blank=True)
    cancellation_details.help_text = "Why was the event cancelled?"

    # Note title is not displayed
    content_panels = [
        PageChooserPanel('overrides'),
        ExceptionDatePanel('except_date'),
        MultiFieldPanel([
            FieldPanel('cancellation_title', classname="full title"),
            FieldPanel('cancellation_details', classname="full")],
            heading="Cancellation"),
        ]
    promote_panels = []

    @property
    def status(self):
        return "cancelled"

    @property
    def status_text(self):
        return "This event has been cancelled."

# ------------------------------------------------------------------------------
class PostponementPageForm(EventExceptionPageForm):
    slugName = "postponement"

    def clean(self):
        cleaned_data = super().clean()
        self._checkSlugAvailable(cleaned_data)
        self._checkSlugAvailable(cleaned_data, "cancellation")
        EventPageForm._checkStartBeforeEnd(self, cleaned_data)
        return cleaned_data

class PostponementPage(EventBase, CancellationPage):
    class Meta:
        verbose_name = "Postponement"

    parent_page_types = ["joyous.RecurringEventPage"]
    subpage_types = []
    base_form_class = PostponementPageForm

    postponement_title = models.CharField('Title', max_length=255)
    postponement_title.help_text = "The title for the postponed event"
    date    = models.DateField("Date")

    content_panels = [
        PageChooserPanel('overrides'),
        ExceptionDatePanel('except_date'),
        MultiFieldPanel([
            FieldPanel('cancellation_title', classname="full title"),
            FieldPanel('cancellation_details', classname="full")],
            heading="Cancellation"),
        MultiFieldPanel([
            FieldPanel('postponement_title', classname="full title"),
            ImageChooserPanel('image'),
            FieldPanel('date'),
            FieldPanel('time_from', widget=TimeInput),
            FieldPanel('time_to', widget=TimeInput),
            FieldPanel('details', classname="full"),
            MapFieldPanel('location'),
            FieldPanel('website')],
            heading="Postponed to"),
    ]
    promote_panels = []

    @classmethod
    def getEventsByDay(cls, date_from, date_to):
        ordFrom =  date_from.toordinal()
        ordTo   =  date_to.toordinal()
        # TODO As with simple event, think about changing to the same algorithm
        # as used for multiday event
        # better yet how about refactoring it out
        events = [EventsOnDay(dt.date.fromordinal(ord), [], [])
                  for ord in range(ordFrom, ordTo+1)]

        dateRange = (date_from-_1day, date_to+_1day)
        pages = PostponementPage.objects.live().filter(date__range=dateRange)
        for page in pages:
            fromDate = getLocalDate(page.date, page.time_from, page.tz)
            toDate   = getLocalDate(page.date, page.time_to, page.tz)
            dayNum = fromDate.toordinal() - ordFrom
            thisEvent = ThisEvent(page.postponement_title, page)
            if 0 <= dayNum <= ordTo - ordFrom:
                events[dayNum].days_events.append(thisEvent)
            if fromDate != toDate:
                if 0 <= dayNum+1 <= ordTo - ordFrom:
                    events[dayNum+1].continuing_events.append(thisEvent)
        return events

    @property
    def tz(self):
        # use the parent's time zone
        return self.overrides.tz

    @property
    def group(self):
        # use the parent's group
        return self.overrides.group

    @property
    def status(self):
        myNow = timezone.localtime(timezone=self.tz)
        if getAwareDatetime(self.date, self.time_to, self.tz) < myNow:
            return "finished"
        elif getAwareDatetime(self.date, self.time_from, self.tz) < myNow:
            return "started"
        return None

    @property
    def when(self):
        return self._getLocalWhen(self.date)

    @property
    def postponed_from_when(self):
        return self.cancellationpage.when

    @property
    def at(self):
        timeFrom = getLocalTime(self.date, self.time_from, self.tz)
        return timeFormat(timeFrom)

    def _getFromDt(self):
        return getLocalDatetime(self.date, self.time_from, self.tz)


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Test Get Event Functions
# ------------------------------------------------------------------------------
import sys
import datetime as dt
import pytz
import calendar
from django.test import TestCase
from django.contrib.auth.models import User
from django.utils import timezone
from wagtail.core.models import Page
from ls.joyous.recurrence import Recurrence
from ls.joyous.recurrence import DAILY, WEEKLY, MONTHLY, MO, TU, WE, TH, FR, WEEKEND, EVERYDAY
from ls.joyous.models.calendar import CalendarPage
from ls.joyous.models.events import (SimpleEventPage, MultidayEventPage,
        RecurringEventPage, PostponementPage, ExtraInfoPage)
from ls.joyous.models.events import (getAllEventsByDay, getAllEventsByWeek,
        getAllUpcomingEvents, getAllPastEvents, getGroupUpcomingEvents)
from ls.joyous.models.groups import get_group_model
from .testutils import datetimetz

GroupPage = get_group_model()

class TestGetEvents(TestCase):
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@foo.test', 's3cr3t')
        self.calendar = CalendarPage(owner = self.user,
                                     slug  = "events",
                                     title = "Events")
        self.home.add_child(instance=self.calendar)
        self.group = GroupPage(slug = "initech", title = "Initech Corporation")
        self.home.add_child(instance=self.group)

        self.show = SimpleEventPage(owner = self.user,
                                    slug   = "pet-show",
                                    title  = "Pet Show",
                                    date      = dt.date(2013,1,5),
                                    time_from = dt.time(11),
                                    time_to   = dt.time(17,30))
        self.calendar.add_child(instance=self.show)

        self.party = MultidayEventPage(owner = self.user,
                                       slug  = "allnighter",
                                       title = "All Night",
                                       date_from = dt.date(2012,12,31),
                                       date_to   = dt.date(2013,1,1),
                                       time_from = dt.time(23),
                                       time_to   = dt.time(3))
        self.calendar.add_child(instance=self.party)

        self.standup = RecurringEventPage(slug   = "test-meeting",
                                          title  = "Test Meeting",
                                          repeat = Recurrence(dtstart=dt.date(2013,1,1),
                                                              until=dt.date(2013,5,31),
                                                              freq=WEEKLY,
                                                              byweekday=[MO,WE,FR]),
                                          time_from = dt.time(13,30),
                                          time_to   = dt.time(16))
        self.group.add_child(instance=self.standup)

        self.postponement = PostponementPage(owner = self.user,
                                             slug  = "2013-01-09-postponement",
                                             title = "Postponement for Wednesday 16th of October",
                                             overrides = self.standup,
                                             except_date = dt.date(2013,1,16),
                                             cancellation_title   = "Meeting Postponed",
                                             cancellation_details =
                                                 "The meeting has been postponed until tomorrow",
                                             postponement_title   = "A Meeting",
                                             date      = dt.date(2013,1,17),
                                             time_from = dt.time(13),
                                             time_to   = dt.time(16,30),
                                             details   = "Yes a test meeting on a Thursday")
        self.standup.add_child(instance=self.postponement)

    def testGetAllEventsByDay(self):
        events = getAllEventsByDay(dt.date(2013,1,1), dt.date(2013,1,31))
        self.assertEqual(len(events), 31)
        evod = events[0]
        self.assertEqual(evod.date, dt.date(2013,1,1))
        self.assertEqual(len(evod.all_events), 1)
        self.assertEqual(len(evod.days_events), 0)
        self.assertEqual(len(evod.continuing_events), 1)


    def testGetAllEventsByWeek(self):
        weeks = getAllEventsByWeek(2013, 1)
        self.assertEqual(len(weeks), 5)
        self.assertIsNone(weeks[0][0])
        self.assertIsNone(weeks[0][1])
        self.assertIsNone(weeks[4][5])
        self.assertIsNone(weeks[4][6])
        evod = weeks[2][4]
        self.assertEqual(evod.date, dt.date(2013,1,17))
        self.assertEqual(len(evod.days_events), 1)
        self.assertEqual(len(evod.continuing_events), 0)

    def testGetAllUpcomingEvents(self):
        today = timezone.localdate()
        futureEvent = MultidayEventPage(owner = self.user,
                                        slug  = "tomorrow",
                                        title = "Tomorrow's Event",
                                        date_from  = today + dt.timedelta(days=1),
                                        date_to    = today + dt.timedelta(days=3),
                                        time_from  = dt.time(17),
                                        time_to    = dt.time(10,30))
        self.calendar.add_child(instance=futureEvent)
        events = getAllUpcomingEvents(self.home)
        self.assertEqual(len(events), 1)
        title, event = events[0]
        self.assertEqual(title, "Tomorrow's Event")
        self.assertEqual(event.slug, "tomorrow")
        events0 = getAllUpcomingEvents()
        self.assertEqual(len(events0), 1)

    def testGetAllPastEvents(self):
        events = getAllPastEvents()
        self.assertEqual(len(events), 4)
        self.assertEqual(events[0].title, "Test Meeting")
        self.assertEqual(events[1].title, "A Meeting")
        self.assertEqual(events[2].title, "Pet Show")
        self.assertEqual(events[3].title, "All Night")

    def testGetGroupUpcomingEvents(self):
        meeting = RecurringEventPage(owner = self.user,
                                     slug  = "plan-plan",
                                     title = "Planning to Plan",
                                     repeat    = Recurrence(dtstart=dt.date(2018,5,1),
                                                            freq=WEEKLY,
                                                            byweekday=[TU]),
                                     time_from = dt.time(18,30),
                                     time_to   = dt.time(20),
                                     group_page = self.group)
        self.calendar.add_child(instance=meeting)
        memo = ExtraInfoPage(owner = self.user,
                             slug  = "plan-plan-extra-info",
                             title = "Extra Information Planning to Plan",
                             overrides = meeting,
                             except_date = meeting.next_date,
                             extra_title = "Gap Analysis",
                             extra_information = "Analyse your gaps")
        meeting.add_child(instance=memo)
        events = getGroupUpcomingEvents(self.group)
        self.assertEqual(len(events), 2)
        self.assertEqual(events[0].title, "Gap Analysis")
        self.assertEqual(events[1].title, "Planning to Plan")




#---------------------------------------------------------------------------
# Testing infrastructure
#---------------------------------------------------------------------------
import unittest
import datetime as dt
from functools import wraps
from django.utils import timezone


def skipUnlessSetup(attrs):
    if type(attrs) == str:
        attrs = [attrs]
    def decorator(func):
        @wraps(func)
        def test(self, *args, **kwargs):
            for attr in attrs:
                if not hasattr(self, attr):
                    self.skipTest("{} not setup".format(attr))
            return func(self, *args, **kwargs)
        return test
    return decorator


def datetimetz(*args):
    if len(args) < 2:
        return timezone.localtime()
    if type(args[0]) == int:
        datetime = dt.datetime(*args)
    elif type(args[0]) == dt.date:
        datetime = dt.datetime.combine(*args)
    return timezone.make_aware(datetime)
